#############################################################################
# regis.py - Pyvox library for image registration
#
#
# This Python module includes various tools for image registration.
#############################################################################

import sys, os
import stat
from math import sqrt, log
from exceptions import Exception

import pyvox, exim
import optim


class RegistrationError (Exception) :
    def __init__(self, args=None) :
        self.args = args



#--------------------------------------------------------------------------
# correl - Weighted correlation of two arrays
#
# This function computes the weighted Pearson product moment correlation of
# the given arrays.
#
# FIXME: Move to a stat module, or redefine as a parray method.
# FIXME: Is this done correctly for an unbiased estimator?
# FIXME: Add functions for mean, var, covar.
# FIXME: Generalize to operate only over specified axes.
#---------------------------------------------------------------------------

def correl(x, y, w=None) :
    if w :
        sumw   = w.asum()         # Sum of the weights
        sumwx  = x.dot(w)         # Weighted sum of first array x
        sumwy  = y.dot(w)         # Weighted sum of second array y
        sumwxx = x.dot(x, w)      # Weighted sum of x*x
        sumwyy = y.dot(y, w)      # Weighted sum of y*y
        sumwxy = x.dot(y, w)      # Weighted sum of x*y
    else :
        sumw   = x.count()
        sumwx  = x.asum()
        sumwy  = y.asum()
        sumwxx = x.dot(x)
        sumwyy = y.dot(y)
        sumwxy = x.dot(y)
    varx   = sumwxx - sumwx * sumwx / sumw
    vary   = sumwyy - sumwy * sumwy / sumw
    result = sumwxy - sumwx * sumwy / sumw
    result = result / sqrt(varx*vary)
    return result



#---------------------------------------------------------------------------
# info - Information content of an unsigned char image
#
# This function computes the information content, in bits per pixel, of an 
# unsigned char image.  The image may be of any rank and shape.  Zero voxels
# are assumed to be background and are ignored.
#
# FIXME: This would go more sensibly into a stat module, if we had a stat
# module.
#---------------------------------------------------------------------------

def info(img, mask=None) :

    # Compute the histogram and discard zero intensities
    hist = img.histo(mask)
    hist[0] = 0

    # Convert to probabilities
    pmin = 1.0 / hist.asum()
    hist = hist.max(1).double() * pmin

    # Compute information
    bits = hist * hist.max(pmin).log()
    bits = bits.asum() / log(2.0)
    return -bits



#----------------------------------------------------------------------------
# mutinfo - Mutual information of two unsigned char images
#
# This function computes the mutual information of two unsigned char images,
# in bits per pixel.  The two images may be of any rank and shape but must
# be the same rank and shape.  An optional mask of per-voxel weights may
# also be provided, and defaults to 1. 
#
# FIXME: Convert to the standard form array.mutinfo?
#
# FIXME: This would go more sensibly into a stat module, if we had a stat
# module.
#
# FIXME: It would be useful to also accept the arguments bi, mar1, mar2 for
# applications in which the bivariate histogram has already been computed.
# This can be distinguished from img1, img2, mask by the fact that the
# arguments in the latter case must all have the same shape.
#
# ALGORITHM:  The various games below limiting histogram bins to no less than
# one are intended to avoid either dividing by zero or taking the log of zero;
# they should have no effect on the result otherwise.
#
# FIXME: Should zero voxels be excluded, as in the info() function?
#----------------------------------------------------------------------------

def mutinfo(img1, img2, mask=None) :

    # Get the bivariate histogram
    (hist, mar1, mar2) = img1.bihist(img2, mask)
    mar1.reshape([256,1])
    mar2.reshape([1,256])

    # Convert to probabilities
    mar1 = mar1.max(1).double() / mar1.asum()
    mar2 = mar2.max(1).double() / mar2.asum()
    pmin = 1 / hist.asum()
    hist = hist.double() * pmin

    # Compute mutual information
    bits = hist.max(pmin) / mar1.mmul(mar2)
    bits = hist * bits.log()
    return bits.asum() / log(2.0)



#---------------------------------------------------------------------------
# obregis - Objective function for image registration
#
# FIXME: It would be useful to have an objective function that could be
# dynamically upgraded from rigid to affine to spline transforms, for
# registration protocols that use progressively more flexible transforms.
#
# FIXME: Make sure that this works for transforms more general than affine.
# On first glance, it ought to work for any transform which is applied
# uniformly over the whole image, such as polynomial or perspective; but
# it won't handle transforms such as splines which want to optimize
# locally.
#
# FIXME: Make it possible to plug in different metrics rather than having
# them hardwired into this code.  One possible approach is to define a
# generic regmetric class with methods scale() and eval(); a init() class
# might be necessary for efficiency.
#
# FIXME: It might be prudent to use a method setmask() to set the mask,
# in case some initialization needs to be done.
#
# FIXME: Most current image registrators are limited to unsigned char
# images only, or even more specifically to 3D unsigned char images.
#
# FIXME: Allow setting the size of the resampled test image.
#
# FIXME: Provide for an intensity mapping as well.
#
# FIXME: Provide a means for controlling how much progress information is
# printed out, and to what device.  Perhaps an iteration hook which calls
# a user-specified function?
#
# FIXME: Implement additional transforms: similarity (rigid plus uniform
# scale); restricted affine (rigid plus scale along each principal axis).
#---------------------------------------------------------------------------

class obregis (optim.obfunction) :

    # Initialize an instance, including optimization parameters
    # Assumes that the derived class has already defined self.npar and
    # that zero is a reasonable initial guess for all parameters.
    def __init__(self, source, target) :
        optim.obfunction.__init__(self, self.npar)

        # Initialize attributes intended to be set by the user
        self.source = source       # Source image to register
        self.target = target       # Target to register to
        self.scenter = None        # Physical "center" of source image
        self.tcenter = None        # Physical "center" of target image
        self.sspacing = None       # Pixel spacing in source
        self.tspacing = None       # Pixel spacing in target
        self.mask   = None         # Image mask/weights in target coords
        self.metric = "mutinfo"    # Metric to use for similarity
        self.interp = "linear"     # Interpolation type to use
        self.niter  = 10           # Number of iterations

        # Initialize attributes intended to be read by the user
        self.xopt   = self.npar * [0,]   # Init/final optimum found
        self.fopt   = None               # Value at optimum

        # Initialize other attributes
        self.funct  = None
        self.src2   = source     # Shrunk images for multi-scale
        self.tgt2   = target
        self.mask2  = self.mask  # Shrunk mask for multi-scale
        self.tmean  = None       # Norms for correl, if needed
        self.tnorm  = None
        self.shrink = 0
        self.smooth = 0

    # Evaluate discrepancy for given params
    def __call__(self, param) :
        size = self.tgt2.size
	if self.mask2 :
            npix = self.mask2.asum()
        else :
            npix = self.tgt2.count()
        aff  = self.stgt2src(param, self.shrink)
        if self.interp == "linear" :
            test = aff.linear(self.src2, size)
        elif self.interp == "nearest" :
            test = aff.nearest(self.src2, size)
	else :
	    raise Error, "Interpolation %s not recognized" % self.interp
        # FIXME: The following is a hack to support intensity scaling for
        # obaffinegb; either do it right or remove it entirely
        if (0) :    
            lower = param[1]
            upper = param[0] * 255.0 + param[1]
            test = test.uchar(lower, upper)
        if self.metric == "mutinfo" :
            # Mutual information metric, in bits per pixel and
            # inverted so minimization optimizes
            fnew = -mutinfo(test, self.tgt2, self.mask2)
        elif self.metric == "correl" :
            # Correlation between two images, scaled to 0,1 range and
            # inverted so that minimization optimizes.
            # FIXME: Arrange to precompute the target mean and variance
            # Compute averages and norms to scale the correlation
            #if not self.tnorm :
            #    self.tnorm = self.tgt2.norm2(0, self.mask2)
            #    self.tmean = self.tgt2.dot(self.mask2) / npix
            #ntst = test.norm2(0, self.mask2)
            #if self.mask2 :
            #    mtst = test.dot(self.mask2) / npix
            #else :
            #    mtst = test.asum() / npix
            # Compute the Pearson product moment correlation
            #fnew = test.dot(self.tgt2, self.mask2) - mtst*self.tmean
            #fnew = -fnew / (ntst * self.tnorm)
            fnew = -correl(test, self.tgt2, self.mask2)
        elif self.metric == "norm2" :
            # L2 metric (akin to ANOVA), expressed per pixel
            fnew = test.norm2(self.tgt2, self.mask2)
            fnew = fnew / sqrt(npix)
        else :
            raise RegistrationError, "Metric %s not recognized" % self.metric
        print "Error at",
        for i in range(len(param)) :
            print "%8.4f" % param[i],
        print "is %7.4f" % fnew,
        if not self.fopt or fnew < self.fopt :
            self.xopt = param
            self.fopt = fnew
            print " *"
        else :
            print
        return fnew

    # Set scale and smoothing levels for test image
    # FIXME: Preserve old shrunk images if possible
    # FIXME: For mutinfo metric, do quantization rather than smoothing;
    #    find cleaner algorithm for shrinking (e.g. image.shrink()!)
    # FIXME: Choose a better control parameter than metric == 'mutinfo'
    #    to decide whether to use linear or NN interpolation in shrink
    def scale(self, shrink, smooth = 0) :
        self.tmean = None     # Forget summary value for old shrink
        self.tnorm = None
        self.src2  = self.source
        self.tgt2  = self.target
        self.mask2 = self.mask
        if self.metric == 'mutinfo' :
            sf = 2.0**shrink
            aff = pyvox.affine(self.ndim)
            aff.scale(sf)
            ssz = self.source.size
            tsz = self.target.size
            for i in range(self.ndim) :
                ssz[i] = ssz[i] / sf
                tsz[i] = tsz[i] / sf
            self.src2 = aff.nearest(self.source, ssz)
            self.tgt2 = aff.nearest(self.target, tsz)
            if self.mask :
                self.mask2 = aff.nearest(self.mask, tsz)
        else :
            self.src2 = self.source
            self.tgt2 = self.target
            for i in range(shrink) :
                self.src2 = self.src2.lowpass(2)
                self.tgt2 = self.tgt2.lowpass(2)
                if self.mask2 :
                    self.mask2 = self.mask2.lowpass(2)
            for i in range(smooth) :
                self.src2 = self.src2.lowpass()
                self.tgt2 = self.tgt2.lowpass()
                # Assume mask does not need to be smoothed
        self.shrink = shrink
        self.smooth = smooth
        self.fopt = None



#---------------------------------------------------------------------------
# obaffine - Objective function for affine registration
#
# Parameters consist of N*N matrix elements, and N translations, where N
# is the number of image dimensions.  A gain or bias difference between
# the two images should not matter for either correl or mutinfo metric.
#
# This class is specialized below to define rigid and translation-only
# transforms.  The methods that must ordinarily be redefined are __init__
# and ctgt2src.
#
# FIXME: Currently supports unsigned char images only.
#
# FIXME: Need a way to scale an uchar image by a float value.  Array.uchar
# with limits works but seems sort of inelegant.
#
# FIXME: Provide a method to print out the parameters with labels for
# human consumption?  Make the stepwise printouts optional?
#--------------------------------------------------------------------------

class obaffine (obregis) :

    # Initialize instance, including optimization parameters
    def __init__(self, source, target, init=None) :
        if source.rank != target.rank :
            raise RegistrationError, \
                  "source and target have different number of dimensions"
        ndim = source.rank
        self.ndim = ndim
        self.npar = ndim*(ndim+1)
        obregis.__init__(self, source, target)
        if init == None or isinstance(init, pyvox.AffineType) :
            self.init = init
        elif isinstance(init, obrigid) or isinstance(init, obaffine) :
            self.init = init.ctgt2src()
            self.metric  = init.metric
            self.scenter = init.scenter
            self.tcenter = init.tcenter
        else :
            raise RegistrationError, \
                  "Cannot translate init argument to an initial transform"
        for i in range(ndim):
            self.xopt[i*(ndim+1)] = 1.0
        self.xtol = ndim*ndim * [0.001,] + ndim * [0.1,]
        self.step = ndim*ndim * [0.01,]  + ndim * [1.0,]
        self.ftol = 0.001       # Assumes per-voxel discrepancy

    # Initialize transform from an obrigid precursor
    # FIXME: This method is ugly!  Consider defining methods set_itgt2src()
    # and set_ptgt2src() instead; then only the names are ugly.
    # FIXME: Remove this method and use the init argument instead?
    def initrigid(self, rigid) :
        self.init     = rigid.ctgt2src()
        self.sspacing = rigid.sspacing
        self.scenter  = rigid.scenter
        self.tspacing = rigid.tspacing
        self.tcenter  = rigid.tcenter

    # Construct centered target-to-source transform; this is the
    # transform directly derived from the transform parameters and
    # thus the form on which the optimizer works.  It is also the
    # affine transform that should be used to initialize a more
    # general type of transform.
    def ctgt2src(self, param=None) :
        if not param :
            param = self.xopt
        if self.init :
            aff = self.init.copy()
        else :
            aff = pyvox.affine(self.ndim)
        n0 = self.ndim
        n1 = n0*n0
        n2 = self.npar
        mat = pyvox.array([n0,n0], exim.double, param[ 0:n1].list())
        off = pyvox.array([n0, 1], exim.double, param[n1:n2].list())
        aff = aff.compose2(mat, off)
        return aff

    # Construct physical target-to-source transform
    def ptgt2src(self, param=None) :
        aff = self.ctgt2src(param)
        if self.tcenter :
            aff.translate(self.tcenter, -1)
        if self.scenter :
            aff.translate(self.scenter)
        return aff

    # Construct index target-to-source transform
    def itgt2src(self, param=None) :
        aff = self.ptgt2src(param)
        if self.tspacing :
            aff.scale(self.tspacing, 1)
        elif self.target.spacing :
            aff.scale(self.target.spacing, 1)
        if self.sspacing :
            aff.scale(self.sspacing, -2)
        elif self.source.spacing :
            aff.scale(self.source.spacing, -2)
        return aff

    # Construct shrunk target-to-source transform
    # For internal use; maps source onto shrunk target image
    def stgt2src(self, param=None, shrink=None) :
        aff = self.itgt2src(param)
        if shrink :
            aff.scale(2.0**shrink, 1)
            aff.scale(1.0/(2.0**shrink))
        return aff

    # Set parameters from an affine transform
    # FIXME: Deprecated
    def setaffine (self, aff) :
        self.xopt = aff.matrix.list() + aff.offset.list()

    # Construct target-to-source transform for given param values
    # FIXME: Deprecated
    def tgt2src(self, param=None, shrink=None) :
        raise Exception, \
              'tgt2src method is deprecated; use itgt2src or ptgt2src instead'
        if not param :
            param = self.xopt
        aff = pyvox.affine(self.ndim)
        if shrink :
            aff.scale(2.0**shrink)
        if self.tcenter :
            aff.translate(-self.tcenter)
        n0 = self.ndim
        n1 = n0*n0
        n2 = self.npar
        mat = pyvox.array([n0,n0], exim.double, param[ 0:n1].list())
        off = pyvox.array([n0, 1], exim.double, param[n1:n2].list())
        aff.compose2(mat, off)
        if self.scenter :
            aff.translate(self.scenter)
        if shrink :
            aff.scale(1.0/(2.0**shrink))
        # aff.write()   # DEBUG
        return aff



#----------------------------------------------------------------------------
# obtrans - Objective function for registration by translation only
#
# The parameters to be optimized consist of one translation for each
# dimension of the source and target.
#
# FIXME: This class has not been tracking changes to the obaffine and
# obrigid classes and may be obsolete now.
#----------------------------------------------------------------------------

class obtrans (obaffine) :

    # Initialize instance, including optimization parameters
    def __init__(self, source, target) :
        if source.rank != target.rank :
            raise RegistrationError, "Source and target have different rank"
        self.ndim = source.rank    # Image dimension
        self.npar = self.ndim      # Nmbr of params to optimize
        obregis.__init__(self, source, target)
        self.xtol   = self.ndim * [ 0.1, ]
        self.ftol   = 0.001      # Assumes per-voxel discrepancy
        self.step   = self.ndim * [ 1.0, ]

    # Construct centered target-to-source transform; this is the
    # transform directly derived from the transform parameters and
    # thus the form on which the optimizer works.  It is also the
    # affine transform that should be used to initialize a more
    # general type of transform.
    def ctgt2src(self, param=None) :
        if not param :
            param = self.xopt
        aff = pyvox.affine(self.ndim)
        aff.translate(param)
        return aff



#----------------------------------------------------------------------------
# obrigid - Objective function for rigid registration
#
# The parameters to be optimized consist of three translations and three
# rotation angles.
#
# FIXME: Currently works for 3D unsigned char only.
#
# FIXME: Risk of gimbal lock if angles get near pi/2
#----------------------------------------------------------------------------

class obrigid (obaffine) :

    # Initialize instance, including optimization parameters
    # FIXME: It seems rather awkward that this class is derived from
    # obaffine but calls obregis.__init__ for initialization.
    def __init__(self, source, target, init=None) :
        if source.rank != 3 or target.rank != 3 :
            raise RegistrationError, "obrigid supports only 3D images"
        self.ndim = 3         # Image dimension
        self.npar = 6         # Nmbr of params to optimize
        obregis.__init__(self, source, target)
        if init == None or isinstance(init, pyvox.AffineType) :
            self.init = init
        elif isinstance(init, obrigid) or isinstance(init, obaffine) :
            self.init = init.ctgt2src()
            self.metric  = init.metric
            self.scenter = init.scenter
            self.tcenter = init.tcenter
        else :
            raise RegistrationError, \
                  "Cannot translate init argument to an initial transform"
        self.step   = [ 1.0, 1.0, 1.0, 0.01, 0.01, 0.01 ]
        self.xtol   = [ 0.1, 0.1, 0.1, 0.001, 0.001, 0.001 ]
        self.ftol   = 0.001      # Assumes per-voxel discrepancy

    # Construct centered target-to-source transform
    # Use this one to initialize an affine registration
    def ctgt2src(self, param=None) :
        if not param :
            param = self.xopt
        if self.init :
            aff = self.init.copy()
        else :
            aff = pyvox.affine(self.ndim)
        aff.translate(param[0:3])
        aff.rotate(1, 2, param[3])
        aff.rotate(0, 2, param[4])
        aff.rotate(0, 1, param[5])
        return aff

    # Construct target-to-source transform for given param values
    # FIXME: Deprecated (because wrong or difficult to use)
    def tgt2src(self, param=None, shrink=None) :
        raise Exception, \
              'tgt2src method is deprecated; use itgt2src or ptgt2src instead'
        if not param :
            param = self.xopt
        aff = pyvox.affine(self.ndim)
        if shrink :
            aff.scale(2.0**shrink)
        if self.tcenter :
            aff.translate(-self.tcenter)
        aff.translate(param[0:3])
        aff.rotate(1, 2, param[3])
        aff.rotate(0, 2, param[4])
        aff.rotate(0, 1, param[5])
        if self.scenter :
            aff.translate(self.scenter)
        if shrink :
            aff.scale(1.0/(2.0**shrink))
        return aff



#---------------------------------------------------------------------------
# obaffinegb - Objective function for affine registration, plus gain and bias
#
# Parameters consist of 2 intensity gain/offsets, 9 matrix elements, and 3 
# translations.
#
# FIXME: This class is arguably obsolete and should be deleted.  If not
# deleted, it needs to be made consistent with obaffine.
#
# FIXME: Currently for 3D unsigned char images only.
#
# FIXME: Need a way to scale an uchar image by a float value.  Array.uchar
# with limits works but seem sort of inelegant.
#
# FIXME: Provide a method to print out the parameters with labels for
# human consumption?  Make the stepwise printouts optional?
#--------------------------------------------------------------------------

class obaffinegb (obregis) :

    # Initialize instance, including optimization parameters
    def __init__(self, source, target) :
        if source.rank != 3 or target.rank != 3 :
            raise RegistrationError, "obaffine supports only 3D images"
        self.ndim =  3         # Image dimension
        self.npar = 14         # Nmbr of params to optimize
        obregis.__init__(self, source, target)
        self.step = [0.01, 5.0] + 9 * [0.01,]  + 3 * [1.0,]
        self.xtol = [0.05, 1.0] + 9 * [0.001,] + 3 * [0.1,]
        self.ftol = 0.05       # Assumes per-voxel discrepancy

    # Set parameters from an affine transform
    def setaffine (self, aff) :
        self.xopt = [1.0, 0.0] + aff.matrix.list() + aff.offset.list()

    # Construct target-to-source transform for given param values
    def tgt2src(self, param=None, shrink=None) :
        if not param :
            param = self.xopt
        aff = pyvox.affine(self.ndim)
        if shrink :
            aff.scale(2.0**shrink)
        if self.tcenter :
            aff.translate(-self.tcenter)
        mat = pyvox.array([3,3], exim.double, param[ 2:11].list())
        off = pyvox.array([3,1], exim.double, param[11:14].list())
        aff.compose2(mat, off)
        if self.scenter :
            aff.translate(self.scenter)
        if shrink :
            aff.scale(1.0/(2.0**shrink))
        return aff
